/*
 * Copyright (C) 2025 RoyR. All rights reserved.
 * Distributed under the terms of the MIT License.
 *
 * Very Simple HTTP server
 * Version 1.0.0
 *   _   _   _   _   _   _
 *  / \ / \ / \ / \ / \ / \
 * ( v | s | H | T | T | D | (Very Simple HTTP Daemon)
 *  \_/ \_/ \_/ \_/ \_/ \_/
 *
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <pwd.h>
#include <libgen.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define BUFFER_SIZE 4096
#define VSHTTPD_VERSION "1.0.0"

static int client_fd;
static int log_enable = 0;
static char *main_docroot;
static short port = 8080;
static char *post_buffer = NULL;
static size_t post_buffer_length;
static size_t post_length = 0;
static char post_type[80+1];

#ifndef IPV6
static struct sockaddr_in client_addr;
#else /* IPV6 is defined */
static struct sockaddr_in6 client_addr;
#endif

static char *getfile_type(char *filename) {
  int filename_length = strlen(filename);
  char *ext;
  while(filename_length > 0 && !(filename[filename_length] == '.')) {
    filename_length--;
  }
  /* fprintf(stderr, "%s\n", filename + filename_length); */
  ext = filename + filename_length;
  if(!strcmp(ext, ".html")) {
    return "text/html";
  } else if(!strcmp(ext, ".txt")) {
    return "text/plain";
  } else if(!strcmp(ext, ".jpg") || !strcmp(ext, ".jpe") || !strcmp(ext, ".jpeg")) {
    return "image/jpeg";
  } else if(!strcmp(ext, ".png")) {
    return "image/png";
  } else {
    return "binary/unknown";
  }
}

static void send_error(FILE *out, int id) {
  char *error_msg;
  char *http_msg = "HTTP/1.1";
  if(id == 404) { /* Not Found */
    fprintf(out, "%s 404 Not Found\r\n", http_msg);
    error_msg = "<h1>404 Not Found</h1>";
  } else if(id == 403) { /* Forbidden */
    fprintf(out, "%s 403 Forbidden\r\n", http_msg);
    error_msg = "<h1>403 Forbidden</h1>";
  } else if(id == 400) {
    fprintf(out, "%s 400 Bad Request\r\n", http_msg);
    error_msg = "<h1>400 Bad Request</h1>";
  } else if(id == 501) {
    fprintf(out, "%s 501 Method Not Implemented\r\n", http_msg);
    error_msg = "<h1>Method Not Implemented</h1>";
  } else {
    fprintf(out, "%s 500 Internal Server Error\r\n", http_msg);
    error_msg = "<h1>500 Internal Server Error\r\n";
  }
  fprintf(out, "Server: vshttpd\r\n");
  fprintf(out, "Content-Type: text/html\r\n");
  fprintf(out, "\r\n");
  fprintf(out, error_msg);
}

static void send_200(FILE *out, char *content_type, off_t size) {
  fprintf(out, "HTTP/1.1 200 OK\r\n");
  fprintf(out, "Server: vshttpd\r\n");
  fprintf(out, "Connection: close\r\n");
  fprintf(out, "Content-Length: %llu\r\n", (uint64_t)size);
  fprintf(out, "Content-Type: %s\r\n\r\n", content_type);
}

static char *build_env(char const *name, char const *value) {
  char *ptr = malloc((strlen(name) + strlen(value)) + 2);

  if(ptr) {
    strcpy(ptr, name);
    strcat(ptr, "=");
    strcat(ptr, value);
  } else {
    perror("Failed to allocate memory for environment variable");
    exit(1);
  }
  return ptr;
}

static void format_ip(void *sockaddr_ptr, char *ip_addr) {
  const char *buf_ptr;
  char buf[50];
  
#ifndef IPV6
  struct in_addr *addr;
  struct sockaddr_in *sockaddr;
  int af = AF_INET;
  
  sockaddr = (struct sockaddr_in *)sockaddr_ptr;
  addr = &sockaddr->sin_addr;
#else
  struct in6_addr *addr;
  struct sockaddr_in6 *sockaddr;
  int af = AF_INET6;
  
  sockaddr = (struct sockaddr_in6 *)sockaddr_ptr;
  addr = &sockaddr->sin6_addr;
#endif
  buf_ptr = inet_ntop(af, (void *)addr, buf, sizeof(buf));
  
  if(buf_ptr) {
    strcpy(ip_addr, buf_ptr);
  }
  return;
}

static void execute_cgi(FILE *out, char *uri, const char *method, const char *doc_root) {
  char *argv[256+1];
  char *envp[15+1] = { NULL };
  int env_num = 0;
  FILE *cgi_out;
  FILE *cgi_in;
  char buffer[BUFSIZ];
  int cgi_out_pipe[2];
  int cgi_in_pipe[2];
  int bytes_read, bytes_written;
  char exec_path[4096];
  char cgi_bin[4096];
  char ip_addr[50];
  char *server_name = basename((char *)doc_root);
  char port_str[5+1];

  if(uri[strlen(uri) - 1] == '/') {
    send_error(out, 403);
    return;
  }
  if((pipe(cgi_out_pipe) < 0) || (pipe(cgi_in_pipe) < 0)) {
    send_error(out, 500);
    perror("pipe");
    return;
  }

  sprintf(port_str, "%hu", port);
  
  buffer[0] = '\0';
  strncat(buffer, uri, strcspn(uri, "?"));

  strcpy(cgi_bin, doc_root);
  strcat(cgi_bin, "/cgi-bin/");
  format_ip(&client_addr, ip_addr);
  envp[env_num] = build_env("DOCUMENT_ROOT", doc_root); env_num++;
  envp[env_num] = build_env("GATEWAY_INTERFACE", "CGI/1.1"); env_num++;
  envp[env_num] = build_env("REQUEST_URI", uri); env_num++;
  envp[env_num] = build_env("SERVER_SOFTWARE", "vshttpd (Unix)"); env_num++;
  envp[env_num] = build_env("HTTP_HOST", server_name); env_num++;
  envp[env_num] = build_env("QUERY_STRING", uri + strcspn(uri, "?") + 1); env_num++;
  envp[env_num] = build_env("REQUEST_METHOD", method); env_num++;
  envp[env_num] = build_env("REMOTE_ADDR", ip_addr); env_num++;
  envp[env_num] = build_env("SCRIPT_NAME", buffer); env_num++;
  envp[env_num] = build_env("SERVER_NAME", server_name); env_num++;
  envp[env_num] = build_env("SERVER_PORT", port_str); env_num++;
  envp[env_num] = build_env("SERVER_PROTOCOL", "HTTP/1.1"); env_num++;
  if(!strcmp(method, "POST")) {
    char content_length[25];
    sprintf(content_length, "%llu", (uint64_t)post_length);
    envp[env_num] = build_env("CONTENT_LENGTH", content_length); env_num++;
    envp[env_num] = build_env("CONTENT_TYPE", post_type); env_num++;
  }
  envp[env_num] = NULL;

  if(strchr(uri, '?')) {
    strcpy(exec_path, doc_root);
    strncat(exec_path, uri, strcspn(uri, "?"));
    int counter = 0 + strcspn(uri, "?");
    int argv_index = 1;
    argv[0] = exec_path;
    while(uri[counter] != 0 && counter < 256) {
      if(uri[counter] == '&' || uri[counter] == '?') {
    	argv[argv_index] = &uri[counter+1];
	uri[counter] = 0;
    	argv_index++;
      }
      counter++;
    }
  } else {
    strcpy(exec_path, doc_root);
    strcat(exec_path, uri);
  }
  if(log_enable)
    fprintf(stderr, "Running CGI: %s, for virtual host: %s\n", exec_path, \
	    basename((char *)doc_root));
  pid_t pid = fork();
  if (pid == 0) {
    /* Child process */
    dup2(cgi_out_pipe[1], STDOUT_FILENO);/* Redirect stdout to the pipe */
    dup2(cgi_in_pipe[0], STDIN_FILENO); /* Redirect stdin from the in pipe */
    close(cgi_in_pipe[1]);
    close(cgi_out_pipe[0]);
    chdir(cgi_bin);
    execve(exec_path, argv, envp); /* Execute the CGI script */
    perror(exec_path);
    exit(0); /* Only reached if there is a problem with th execv */
  } else if(pid > 0) {
    /* Parent process */
    env_num = 0;
    while(envp[env_num]) {
      free(envp[env_num]);
      env_num++;
    }
    cgi_out = fdopen(cgi_out_pipe[0], "r");
    cgi_in = fdopen(cgi_in_pipe[1], "w");
    close(cgi_in_pipe[0]); /* Close the unneed ends of the pipes */
    close(cgi_out_pipe[1]);

    if(!strcmp(method, "POST")) {
      bytes_read = 0;
      if(post_buffer_length > fwrite(post_buffer, 1, \
				     post_buffer_length, cgi_in)) {
	send_error(out, 500);
	fclose(cgi_out);
	close(cgi_out_pipe[0]);
	close(cgi_in_pipe[1]);
	fclose(cgi_in);
	return;
      }
      size_t byte_counter = 0;
      if(post_buffer_length < post_length) {
	bytes_written = 1;
	while(byte_counter < (post_length - post_buffer_length)) {
	  bytes_read = recv(client_fd, buffer, sizeof(buffer), 0);
	  if(bytes_read > 0 && bytes_written > 0) {
	    bytes_written = fwrite(buffer, 1, bytes_read, cgi_in);
	    byte_counter += bytes_read;
	  } else {
	    break;
	  }
	}
      }
    }
    fflush(cgi_in);
    close(cgi_in_pipe[1]); /* Close cgi_in_pipe */
    fclose(cgi_in);
    char content_type[BUFSIZ];
    content_type[0] = 0;
    while(fgets(buffer, BUFSIZ, cgi_out) != NULL) {
      if((strstr(buffer, "Content-Type: ") == buffer) ||
	 (strstr(buffer, "Content-type: ") == buffer)) {
	strncpy(content_type, buffer, sizeof(content_type));
      }
      if((strstr(buffer, "\r\n") == buffer) ||
	 (strstr(buffer, "\n") == buffer))
	break;
    }
    if(!strlen(content_type)) {
      send_error(out, 500);
      fclose(cgi_out);
      close(cgi_out_pipe[0]);
      return;
    }
    fprintf(out, "HTTP/1.1 200 OK\r\n");
    fprintf(out, "Server: vshttpd\r\n");
    fprintf(out, content_type);
    fprintf(out, buffer);
    
    if(strcmp(method, "HEAD")) {
      bytes_read = fread(&buffer, 1, sizeof(buffer), cgi_out);
      bytes_written = 1;
      while (bytes_read > 0 && bytes_written > 0) {
	bytes_written = fwrite(&buffer, 1, bytes_read, out);
	bytes_read = fread(&buffer, 1, sizeof(buffer), cgi_out);
      }
    }
    fclose(cgi_out);
    close(cgi_out_pipe[0]);
    wait(NULL); /* Wait for the child to exit */
  } else {
    /* fork() failed */
    perror("fork");
    send_error(out, 500);
    env_num = 0;
    while(envp[env_num]) {
      free(envp[env_num]);
      env_num++;
    }
  }
}

static void serve_file(FILE *out, const char *filename, const char *method) {
  FILE *file = fopen(filename, "rb");
  struct stat file_stats;
  size_t bytes_read;
  size_t bytes_writen;

  if(!file || stat(filename, &file_stats)) {
    if(access(filename, F_OK) < 0) {
      send_error(out, 404);
    } else if(access(filename, R_OK) < 0) {
      send_error(out, 403);
    }
    return;
  }

  send_200(out, getfile_type((char *)filename), file_stats.st_size);
  if(strcmp(method, "HEAD")) {
    char buffer[BUFFER_SIZE];
    bytes_read = fread(&buffer, 1, sizeof(buffer), file);
    bytes_writen = 1;
    while (bytes_read > 0 && bytes_writen > 0) {
      bytes_writen = fwrite(buffer, 1, bytes_read, out);
      bytes_read = fread(buffer, 1, sizeof(buffer), file);
    }
  }
  fclose(file);
}

static void handle_request(FILE *out, const char *host, const char *request) {
  char method[8+1];
  char uri[2048+1];
  char path[4096+1];
  char docroot[4096+1];
  char ip_addr[50+1];

  if(log_enable)
    format_ip(&client_addr, ip_addr);
  
  sscanf(request, "%8s %2048s", method, uri);
  if(strstr(host, "/..") || strstr(uri, "/..")) {
    send_error(out, 400);
    return;
  }
  strcpy(path, main_docroot);
  strcat(path, "/");
  strcat(path, host);
  strcpy(docroot, path);
  strcat(path, uri);
  /* Simple virtual hosting based on the hostname */
  if(!access(docroot, F_OK)) {
    if(access(docroot, R_OK) < 0) {
      /* 403 Forbidden */
      send_error(out, 403);
      return;
    }
    if(log_enable)
      fprintf(stderr, "%s, %s %s for virtual host: %s\n", ip_addr, method, \
	      uri, host);
    if(strcmp(method, "GET") && strcmp(method, "HEAD") && \
       strcmp(method, "POST")) {
      send_error(out, 501);
      return;
    }
    if(strstr(uri, "/cgi-bin/") == uri) {
      execute_cgi(out, uri, method, docroot);
    } else if(strcmp(method, "POST")) {
      if(path[strlen(path)-1] == '/') {
	strcat(path, "index.html");
      }
      serve_file(out, path, method);
    } else {
      send_error(out, 400);
    }
  } else {
    send_error(out, 404);
  }
}

static void handle_client(int client_fd) {
  char buffer[BUFFER_SIZE];
  FILE *out = fdopen(client_fd, "w");

  ssize_t buffer_index = 0;
  ssize_t bytes_read = 0;

  memset(buffer, 0, sizeof(buffer));
  bytes_read = recv(client_fd, buffer, sizeof(buffer) - 1, 0);
  buffer_index = bytes_read;
  if(buffer_index > 0) {
    while(!strstr(buffer, "\r\n\r\n") && !strstr(buffer, "\n\n") && \
	  (sizeof(buffer) - buffer_index) - 1 > 1) {
      bytes_read = recv(client_fd, buffer + buffer_index, sizeof(buffer) - \
			buffer_index - 1, 0);
      if(bytes_read > 0) {
	buffer_index += bytes_read;
      } else {
	break;
      }
    }
    char host[256+1];
    host[0] = '\0';
    char *host_header = strstr(buffer, "Host: ");
    if(host_header) {
      sscanf(host_header + 6, "%256s", host);
    } else {
      send_error(out, 400);
      fclose(out);
      return;
    }
    if(strstr(buffer, "POST ") == buffer) {
      char *content_length_header = strstr(buffer, "Content-Length: ");
      if(!content_length_header)
	content_length_header = strstr(buffer, "Content-length: ");
      char *content_type_header = strstr(buffer, "Content-Type: ");
      if(!content_type_header)
	content_type_header = strstr(buffer, "Content-type: ");
      if(content_type_header)
	sscanf(content_type_header + 14, "%80s", post_type);
      else
	strcpy(post_type, "binary/unkown");

      size_t content_length = 0;
      if(content_length_header) {
	sscanf(content_length_header + 16, "%llu", \
	       (uint64_t *)&content_length);
      } else {
	send_error(out, 400);
	fclose(out);
	return;
      }
      char *headers_end;
      headers_end = strstr(buffer, "\r\n\r\n");
      if(!headers_end) {
	headers_end = strstr(buffer, "\n\n"); headers_end += 2;
      } else {
	headers_end += 4;
      }
      if(headers_end && (headers_end < buffer + buffer_index)) {
	post_buffer = headers_end;
	post_buffer_length = (buffer + buffer_index) - headers_end;
      }
      post_length = content_length;
    }
    handle_request(out, host, buffer);
  }
  fclose(out);
  close(client_fd);
}

static void sigpipe_handler(int sig) {
  close(client_fd);
  exit(sig);
}

static void usage(char *name) {
  fprintf(stderr, "Usage: %s [-hlpv] [-d docroot] [-p port] [-u user]\n", name);
  fprintf(stderr, "-d docroot\tSet the document root (default /var/www/html)\n");
  fprintf(stderr, "-h\t\tShow this help\n");
  fprintf(stderr, "-l\t\tEnable logging to STDERR\n");
  fprintf(stderr, "-p port\t\tPort number to listen on (default 8080)\n");
  fprintf(stderr, "-u user\t\tUser to switch to if run by root (default httpd)\n");
  fprintf(stderr, "-v\t\tShow copyright and version infomation\n");
  return;
}

static void version(void) {
  fprintf(stderr, "vshttpd " VSHTTPD_VERSION "\n");
  fprintf(stderr, "Copyright (C) 2024, 2025 RoyR. All rights reserved.\n");
  fprintf(stderr, "Distributed under the terms of the MIT License.\n");
  return;
}

int main(int argc, char *argv[]) {
  int server_fd;
  int optval = 1;
#ifndef IPV6
  struct sockaddr_in server_addr;
  short af = AF_INET;
#else /* IPV6 is defined */
  struct sockaddr_in6 server_addr;
  short af = AF_INET6;
#endif
  int opt;
  char *user = "httpd";
  main_docroot = "/var/www/html";
  extern char *optarg;

  while((opt = getopt(argc, argv, "hd:lp:u:v")) != -1) {
    switch(opt) {
    case 'l':
      log_enable = 1;
      break;
    case 'p':
      if(!(port = atoi(optarg))) {
	fprintf(stderr, "port is an integer in the range 1-65536\n");
	return 1;
      }
      break;
    case 'd':
      main_docroot = optarg;
      break;
    case 'u':
      user = optarg;
      break;
    case 'h':
      usage(argv[0]);
      return 1;
    case 'v':
      version();
      return 1;
    default:
      usage(argv[0]);
      return 1;
    }
  }
  if(optind < argc) {
    usage(argv[0]);
  }

  socklen_t client_len = sizeof(client_addr);
  struct passwd *passwd_info;
  
  /* Create socket */
  if ((server_fd = socket(af, SOCK_STREAM, 0)) < 0) {
    perror("Failed to create socket");
    return 1;
  }
  
  /* Set up server address and bind */
  memset(&server_addr, 0, sizeof(server_addr));
#ifndef IPV6
  server_addr.sin_family = af;
  server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  server_addr.sin_port = htons(port);
#else
  server_addr.sin6_family = af;
  server_addr.sin6_addr = in6addr_any;
  server_addr.sin6_port = htons(port);
#endif
  
  if(setsockopt(server_fd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) != 0) {
    perror("Failed to set socket options");
    return 1;
  }
  
  if(bind(server_fd, (struct sockaddr *)&server_addr, sizeof(server_addr)) < 0) {
    perror("Failed to bind socket");
    return 1;
  }
  
  if(listen(server_fd, 10) > 0) {
    perror("Failed to listen on socket!");
    return 1;
  }
  
  if(getuid() == 0) {
    passwd_info = getpwnam(user);
    if(passwd_info == NULL) {
      fprintf(stderr, "User \"%s\" does not exist\n", user);
      return 1;
    }
    setuid(passwd_info->pw_uid);
    setgid(passwd_info->pw_gid);
  }

  signal(SIGPIPE, sigpipe_handler);
  signal(SIGCHLD, SIG_IGN);

  fprintf(stderr, "Server listening on port %d\n", port);
  
  while(1) {
    client_fd = accept(server_fd, (struct sockaddr *)&client_addr, \
		       &client_len);
    if (client_fd < 0) {
      perror("Failed to accept connection!");
      continue;
    }
    if(fork() == 0) {
      close(server_fd);
      signal(SIGCHLD, SIG_DFL);
      handle_client(client_fd);
      exit(0);
    } else {
      close(client_fd);
    }
  }
  close(server_fd);
  return 0;
}
